@@ -58,6 +58,7 @@
 #include "compression.h"
 #endif
 #include "iso7816.h"
+#include "gp.h"
 #include "../pkcs11/pkcs11.h"
 
 
@@ -88,7 +89,7 @@
 #define COOLKEY_INS_COMPUTE_CRYPT              0x36
 #define COOLKEY_INS_COMPUTE_ECC_KEY_AGREEMENT  0x37
 #define COOLKEY_INS_COMPUTE_ECC_SIGNATURE      0x38
-#define COOLKEY_INS_GET_RANDOM                 0x73
+#define COOLKEY_INS_GET_RANDOM                 0x72
 #define COOLKEY_INS_READ_OBJECT                0x56
 #define COOLKEY_INS_WRITE_OBJECT               0x54
 #define COOLKEY_INS_LOGOUT                     0x61
@@ -1128,7 +1129,7 @@ static int coolkey_read_object(sc_card_t *card, unsigned long object_id, size_t
 	do {
 		ulong2bebytes(&params.offset[0], offset);
 		params.length = MIN(left, COOLKEY_MAX_CHUNK_SIZE);
-		len = left+2;
+		len = left;
 		r = coolkey_apdu_io(card, COOLKEY_CLASS, COOLKEY_INS_READ_OBJECT, 0, 0,
 			(u8 *)&params, sizeof(params), &out_ptr, &len, nonce, nonce_size);
 		if (r < 0) {
@@ -1195,7 +1196,7 @@ static int coolkey_read_binary(sc_card_t *card, unsigned int idx,
 {
 	coolkey_private_data_t * priv = COOLKEY_DATA(card);
 	int r = 0, len;
-	u8 *data = NULL;;
+	u8 *data = NULL;
 
 	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
 	if (idx > priv->obj->length) {
@@ -1466,7 +1467,7 @@ coolkey_find_attribute(sc_card_t *card, sc_cardctl_coolkey_attribute_t *attribut
 	for (i=0; i < attribute_count; i++) {
 		size_t record_len = coolkey_get_attribute_record_len(attr, object_record_type, buf_len);
 		/* make sure we have the complete record */
-		if (buf_len < record_len) {
+		if (buf_len < record_len || record_len < 4) {
 				return SC_ERROR_CORRUPTED_DATA;
 		}
 		/* does the attribute match the one we are looking for */
@@ -1609,40 +1610,17 @@ static int coolkey_card_ctl(sc_card_t *card, unsigned long cmd, void *ptr)
 
 static int coolkey_get_challenge(sc_card_t *card, u8 *rnd, size_t len)
 {
-	size_t rbuflen = 0;
-	int r;
+	LOG_FUNC_CALLED(card->ctx);
 
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
+	if (len > COOLKEY_MAX_CHUNK_SIZE)
+		len = COOLKEY_MAX_CHUNK_SIZE;
 
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-		 "challenge len=%"SC_FORMAT_LEN_SIZE_T"u", len);
-
-	r = sc_lock(card);
-	if (r != SC_SUCCESS)
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
-
-	for(; len >= COOLKEY_MAX_CHUNK_SIZE; len -= COOLKEY_MAX_CHUNK_SIZE,
-										rnd += COOLKEY_MAX_CHUNK_SIZE) {
-		rbuflen = COOLKEY_MAX_CHUNK_SIZE;
-		r = coolkey_apdu_io(card, COOLKEY_CLASS, COOLKEY_INS_GET_RANDOM,
-			0, 0, NULL, 0, &rnd, &rbuflen,  NULL, 0);
-		if (r != COOLKEY_MAX_CHUNK_SIZE) {
-			len = 0;
-			break;
-		}
-	}
-	if (len) {
-		r = coolkey_apdu_io(card, COOLKEY_CLASS, COOLKEY_INS_GET_RANDOM,
-			0, 0, NULL, 0, &rnd, &len, NULL, 0);
-	}
-	sc_unlock(card);
-
-	if (r > 0) {
-		r= SC_SUCCESS;
-	}
-
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+	LOG_TEST_RET(card->ctx,
+			coolkey_apdu_io(card, COOLKEY_CLASS, COOLKEY_INS_GET_RANDOM, 0, 0,
+				NULL, 0, &rnd, &len,  NULL, 0),
+			"Could not get challenge");
 
+	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, (int) len);
 }
 
 static int coolkey_set_security_env(sc_card_t *card, const sc_security_env_t *env, int se_num)
@@ -1951,7 +1929,7 @@ static int coolkey_select_file(sc_card_t *card, const sc_path_t *in_path, sc_fil
 		*file_out = file;
 	}
 
-    return SC_SUCCESS;
+	return SC_SUCCESS;
 }
 
 static int coolkey_finish(sc_card_t *card)
@@ -2144,12 +2122,6 @@ static int coolkey_initialize(sc_card_t *card)
 	if (card->drv_data) {
 		return SC_SUCCESS;
 	}
-
-	/* Select a coolkey read the coolkey objects out */
-	r = coolkey_select_applet(card);
-	if (r < 0) {
-		return r;
-	}
 	sc_debug(card->ctx, SC_LOG_DEBUG_VERBOSE,"Coolkey Applet found");
 
 	priv = coolkey_new_private_data();
@@ -2161,6 +2133,13 @@ static int coolkey_initialize(sc_card_t *card)
 	if (r < 0) {
 		goto cleanup;
 	}
+
+	/* Select a coolkey read the coolkey objects out */
+	r = coolkey_select_applet(card);
+	if (r < 0) {
+		goto cleanup;
+	}
+
 	priv->protocol_version_major = life_cycle.protocol_version_major;
 	priv->protocol_version_minor = life_cycle.protocol_version_minor;
 	priv->pin_count = life_cycle.pin_count;
@@ -2211,6 +2190,14 @@ static int coolkey_initialize(sc_card_t *card)
 	/* if we didn't pull the cuid from the combined object, then grab it now */
 	if (!combined_processed) {
 		global_platform_cplc_data_t cplc_data;
+		/* select the card manager, because a card with applet only will have
+		   already selected the coolkey applet */
+
+		r = gp_select_card_manager(card);
+		if (r < 0) {
+			goto cleanup;
+		}
+
 		r = coolkey_get_cplc_data(card, &cplc_data);
 		if (r < 0) {
 			goto cleanup;
@@ -2259,7 +2246,7 @@ static int coolkey_init(sc_card_t *card)
 
 	r = coolkey_initialize(card);
 	if (r < 0) {
-		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
+		SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INVALID_CARD);
 	}
 
 	card->type = SC_CARD_TYPE_COOLKEY_GENERIC;
