@@ -60,70 +60,72 @@ struct sc_aid sc_hsm_aid = { { 0xE8,0x2B,0x06,0x01,0x04,0x01,0x81,0xC3,0x1F,0x02
 
 
 /* Known ATRs for SmartCard-HSMs */
-static struct sc_atr_table sc_hsm_atrs[] = {
+static const struct sc_atr_table sc_hsm_atrs[] = {
 	/* standard version */
 	{"3B:FE:18:00:00:81:31:FE:45:80:31:81:54:48:53:4D:31:73:80:21:40:81:07:FA", NULL, NULL, SC_CARD_TYPE_SC_HSM, 0, NULL},
 	{"3B:8E:80:01:80:31:81:54:48:53:4D:31:73:80:21:40:81:07:18", NULL, NULL, SC_CARD_TYPE_SC_HSM, 0, NULL},
+	{"3B:DE:18:FF:81:91:FE:1F:C3:80:31:81:54:48:53:4D:31:73:80:21:40:81:07:1C", NULL, NULL, SC_CARD_TYPE_SC_HSM, 0, NULL},
+
 	{"3B:80:80:01:01", NULL, NULL, SC_CARD_TYPE_SC_HSM_SOC, 0, NULL},	// SoC Sample Card
 	{
 		"3B:84:80:01:47:6f:49:44:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:85:80:01:47:6f:49:44:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:86:80:01:47:6f:49:44:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:87:80:01:47:6f:49:44:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:88:80:01:47:6f:49:44:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:89:80:01:47:6f:49:44:00:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:8a:80:01:47:6f:49:44:00:00:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:8b:80:01:47:6f:49:44:00:00:00:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:8c:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:8d:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:8e:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{
 		"3B:8f:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00:00:00:00",
 		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00:00:00:00",
-		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+		"GoID", SC_CARD_TYPE_SC_HSM_GOID, 0, NULL
 	},
 	{NULL, NULL, NULL, 0, 0, NULL}
 };
@@ -190,7 +192,7 @@ static int sc_hsm_select_file_ex(sc_card_t *card,
 		return rv;
 	}
 
-	if ((in_path->value[0] == 0x3F) && (in_path->value[1] == 0x00)) {
+	if ((in_path->len >= 2) && (in_path->value[0] == 0x3F) && (in_path->value[1] == 0x00)) {
 		// The SmartCard-HSM is an applet that is not default selected. Simulate selection of the MF
 		if (in_path->len == 2) {
 			file = sc_file_new();
@@ -225,20 +227,49 @@ static int sc_hsm_select_file(sc_card_t *card,
 
 
 
+static int sc_hsm_get_challenge(struct sc_card *card, unsigned char *rnd, size_t len)
+{
+	LOG_FUNC_CALLED(card->ctx);
+
+	if (len > 1024) {
+		len = 1024;
+	}
+
+	LOG_FUNC_RETURN(card->ctx, iso_ops->get_challenge(card, rnd, len));
+}
+
+
+
 static int sc_hsm_match_card(struct sc_card *card)
 {
 	sc_path_t path;
-	int i, r;
+	int i, r, type = 0;
+	sc_file_t *file = NULL;
 
-	i = _sc_match_atr(card, sc_hsm_atrs, &card->type);
-	if (i >= 0)
+	i = _sc_match_atr(card, sc_hsm_atrs, &type);
+	if (i >= 0 && type != SC_CARD_TYPE_SC_HSM_SOC) {
+		card->type = type;
 		return 1;
+	}
 
 	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
-	r = sc_hsm_select_file(card, &path, NULL);
+	r = sc_hsm_select_file(card, &path, &file);
 	LOG_TEST_RET(card->ctx, r, "Could not select SmartCard-HSM application");
 
-	card->type = SC_CARD_TYPE_SC_HSM;
+	// Validate that card returns a FCP with a proprietary tag 85 with value longer than 2 byte (Fixes #1377)
+	if (file != NULL) {
+		i = file->prop_attr_len;
+		sc_file_free(file);
+		if (i < 2) {
+			return 0;
+		}
+	}
+
+	if (type == SC_CARD_TYPE_SC_HSM_SOC) {
+		card->type = SC_CARD_TYPE_SC_HSM_SOC;
+	} else {
+		card->type = SC_CARD_TYPE_SC_HSM;
+	}
 
 	return 1;
 }
@@ -284,7 +315,9 @@ static int sc_hsm_soc_select_minbioclient(sc_card_t *card)
 	};
 
 	/* Select MinBioClient */
+#ifdef ENABLE_SM
 	sc_sm_stop(card);
+#endif
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xA4, 0x04, 0x0C);
 	apdu.data = minBioClient_aid.value;
 	apdu.datalen = minBioClient_aid.len;
@@ -303,30 +336,62 @@ static int sc_hsm_soc_change(sc_card_t *card, struct sc_pin_cmd_data *data,
 	sc_path_t path;
 	int r;
 
-	/* Select MinBioClient */
-	r = sc_hsm_soc_select_minbioclient(card);
-	LOG_TEST_RET(card->ctx, r, "Could not select MinBioClient application");
+	if (card->type == SC_CARD_TYPE_SC_HSM_SOC) {
+		/* Select MinBioClient */
+		r = sc_hsm_soc_select_minbioclient(card);
+		LOG_TEST_RET(card->ctx, r, "Could not select MinBioClient application");
 
-	/* verify PIN */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0x00, 0x80);
-	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
-	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_GOTO_ERR(card->ctx, r, "Could not verify PIN");
+		/* verify PIN */
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0x00, 0x80);
+		r = sc_transmit_apdu(card, &apdu);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Could not verify PIN");
 
-	/* change PIN */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x24, 0x01, 0x80);
-	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
-	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_GOTO_ERR(card->ctx, r, "Could not change PIN");
+		/* change PIN */
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x24, 0x01, 0x80);
+		r = sc_transmit_apdu(card, &apdu);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Could not change PIN");
+	} else {
+#ifdef ENABLE_SM
+		unsigned sm_mode = card->sm_ctx.sm_mode;
+#endif
+
+		/* verify PIN */
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0x00, 0x85);
+		apdu.cla = 0x80;
+		r = sc_transmit_apdu(card, &apdu);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+
+#ifdef ENABLE_SM
+		/* temporary disable SM, change reference data does not reach the applet */
+		card->sm_ctx.sm_mode = SM_MODE_NONE;
+#endif
+
+		/* change PIN */
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x24, 0x01, 0x85);
+		apdu.cla = 0x80;
+		r = sc_transmit_apdu(card, &apdu);
+#ifdef ENABLE_SM
+		/* restore SM if possible */
+		card->sm_ctx.sm_mode = sm_mode;
+#endif
+		LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+		r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Could not change PIN");
+	}
 
 err:
-	/* Select SC-HSM */
-	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
-	LOG_TEST_RET(card->ctx,
-			sc_hsm_select_file_ex(card, &path, 1, NULL),
-			"Could not select SmartCard-HSM application");
+	if (card->type == SC_CARD_TYPE_SC_HSM_SOC) {
+		/* Select SC-HSM */
+		sc_path_set(&path, SC_PATH_TYPE_DF_NAME,
+			   	sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
+		LOG_TEST_RET(card->ctx,
+				sc_hsm_select_file_ex(card, &path, 1, NULL),
+				"Could not select SmartCard-HSM application");
+	}
 
 	return r;
 }
@@ -338,6 +403,10 @@ static int sc_hsm_soc_unblock(sc_card_t *card, struct sc_pin_cmd_data *data,
 	sc_path_t path;
 	int r;
 
+	if (card->type == SC_CARD_TYPE_SC_HSM_GOID) {
+		return SC_ERROR_NOT_SUPPORTED;
+	}
+
 	/* Select MinBioClient */
 	r = sc_hsm_soc_select_minbioclient(card);
 	LOG_TEST_RET(card->ctx, r, "Could not select MinBioClient application");
@@ -373,19 +442,24 @@ static int sc_hsm_soc_biomatch(sc_card_t *card, struct sc_pin_cmd_data *data,
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
 	int r;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x20, 0x00, 0x85);
-	apdu.cla = 0x80;
-	apdu.data = (unsigned char*)"\x7F\x24\x00";
-	apdu.datalen = 3;
-	apdu.lc = 3;
-	apdu.resplen = 0;
+	if (card->type == SC_CARD_TYPE_SC_HSM_SOC) {
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x20, 0x00, 0x85);
+		apdu.cla = 0x80;
+		apdu.data = (unsigned char*)"\x7F\x24\x00";
+		apdu.datalen = 3;
+		apdu.lc = 3;
+		apdu.resplen = 0;
 
-	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-	/* ignore the actual status bytes */
+		r = sc_transmit_apdu(card, &apdu);
+		LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+		/* ignore the actual status bytes */
+	}
 
 	/* JCOP's SM accelerator is incapable of using case 1 APDU in SM */
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0x20, 0x00, 0x81);
+	if (card->type == SC_CARD_TYPE_SC_HSM_GOID) {
+		apdu.cla = 0x80;
+	}
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof rbuf;
 	r = sc_transmit_apdu(card, &apdu);
@@ -402,6 +476,7 @@ static int sc_hsm_soc_biomatch(sc_card_t *card, struct sc_pin_cmd_data *data,
 
 
 
+#ifdef ENABLE_SM
 #ifdef ENABLE_OPENPACE
 #include "sm/sm-eac.h"
 #include <eac/cv_cert.h>
@@ -498,7 +573,7 @@ static int sc_hsm_perform_chip_authentication(sc_card_t *card)
 		goto err;
 	}
 	EVP_PKEY_free(ctx->ca_ctx->ka_ctx->key);
-	CRYPTO_add(&ctx->ta_ctx->pub_key->references, 1, CRYPTO_LOCK_EVP_PKEY);
+	EVP_PKEY_up_ref(ctx->ta_ctx->pub_key);
 	ctx->ca_ctx->ka_ctx->key = ctx->ta_ctx->pub_key;
 
 	/* generate keys for CA */
@@ -524,6 +599,7 @@ static int sc_hsm_perform_chip_authentication(sc_card_t *card)
 	return SC_ERROR_NOT_SUPPORTED;
 }
 #endif
+#endif
 
 
 
@@ -533,7 +609,9 @@ static int sc_hsm_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
 	sc_apdu_t apdu;
 	u8 cmdbuff[16];
+#ifdef ENABLE_SM
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+#endif
 	int r;
 	int cmd = data->cmd;
 	size_t pin2_len = data->pin2.len;
@@ -558,16 +636,32 @@ static int sc_hsm_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 		return sc_hsm_soc_unblock(card, data, tries_left);
 	}
 
+#ifdef ENABLE_SM
 	/* For contactless cards always establish a secure channel before PIN
 	 * verification. Also, Session PIN generation requires SM. */
-	if ((card->type == SC_CARD_TYPE_SC_HSM_SOC || card->reader->uid.len
-				|| cmd == SC_PIN_CMD_GET_SESSION_PIN)
-			&& (data->cmd != SC_PIN_CMD_GET_INFO)
-			&& card->sm_ctx.sm_mode != SM_MODE_TRANSMIT) {
-		LOG_TEST_RET(card->ctx,
-				sc_hsm_perform_chip_authentication(card),
-				"Could not perform chip authentication");
+	if ((card->type == SC_CARD_TYPE_SC_HSM_SOC
+				|| card->type == SC_CARD_TYPE_SC_HSM_GOID
+				|| card->reader->uid.len || cmd == SC_PIN_CMD_GET_SESSION_PIN)
+			&& (data->cmd != SC_PIN_CMD_GET_INFO)) {
+		struct sc_pin_cmd_data check_sm_pin_data;
+		memset(&check_sm_pin_data, 0, sizeof(check_sm_pin_data));
+		check_sm_pin_data.cmd = SC_PIN_CMD_GET_INFO;
+		check_sm_pin_data.pin_type = data->pin_type;
+		check_sm_pin_data.pin_reference = data->pin_reference;
+
+		r = SC_ERROR_NOT_ALLOWED;
+		if (card->sm_ctx.sm_mode == SM_MODE_TRANSMIT) {
+			/* check if the existing SM channel is still valid */
+			r = sc_pin_cmd(card, &check_sm_pin_data, NULL);
+		}
+		if (r == SC_ERROR_ASN1_OBJECT_NOT_FOUND || r == SC_ERROR_NOT_ALLOWED) {
+			/* need to establish a new SM channel */
+			LOG_TEST_RET(card->ctx,
+					sc_hsm_perform_chip_authentication(card),
+					"Could not perform chip authentication");
+		}
 	}
+#endif
 
 	if ((card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)
 			&& (data->cmd == SC_PIN_CMD_VERIFY)
@@ -604,6 +698,7 @@ static int sc_hsm_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 			data->apdu = &apdu;
 		}
 
+#ifdef ENABLE_SM
 		if ((data->cmd == SC_PIN_CMD_GET_INFO)
 				&& (card->sm_ctx.sm_mode == SM_MODE_TRANSMIT)) {
 			/* JCOP's SM accelerator is incapable of using case 1 APDU in SM */
@@ -612,6 +707,7 @@ static int sc_hsm_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 			apdu.resplen = sizeof rbuf;
 			data->apdu = &apdu;
 		}
+#endif
 
 		data->pin1.offset = 5;
 		data->pin1.length_offset = 4;
@@ -627,11 +723,17 @@ static int sc_hsm_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 		data->cmd = SC_PIN_CMD_GET_SESSION_PIN;
 		if (data->pin_reference == 0x81) {
 			u8 recvbuf[SC_MAX_APDU_BUFFER_SIZE];
+#ifdef ENABLE_SM
 			if (card->sm_ctx.sm_mode != SM_MODE_TRANSMIT) {
 				sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
 						"Session PIN generation only supported in SM");
 				LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 			}
+#else
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+					"Session PIN generation only supported in SM");
+			LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+#endif
 			sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x5A, 0x01, data->pin_reference);
 			apdu.cla = 0x80;
 			apdu.resp = recvbuf;
@@ -669,7 +771,9 @@ static int sc_hsm_logout(sc_card_t * card)
 	sc_path_t path;
 	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
 	memset(priv->sopin, 0, sizeof(priv->sopin));
+#ifdef ENABLE_SM
 	sc_sm_stop(card);
+#endif
 
 	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
 
@@ -746,15 +850,15 @@ static int sc_hsm_write_ef(sc_card_t *card,
 	*p++ = idx & 0xFF;
 	*p++ = 0x53;
 	if (count < 128) {
-		*p++ = count;
+		*p++ = (u8) count;
 		len = 6;
 	} else if (count < 256) {
 		*p++ = 0x81;
-		*p++ = count;
+		*p++ = (u8) count;
 		len = 7;
 	} else {
 		*p++ = 0x82;
-		*p++ = count >> 8;
+		*p++ = (count >> 8) & 0xFF;
 		*p++ = count & 0xFF;
 		len = 8;
 	}
@@ -882,6 +986,8 @@ static int sc_hsm_set_security_env(sc_card_t *card,
 			} else {
 				priv->algorithm = ALGO_RSA_PKCS1;
 			}
+		} else if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PSS) {
+			priv->algorithm = ALGO_RSA_PSS;
 		} else {
 			if (env->operation == SC_SEC_OPERATION_DECIPHER) {
 				priv->algorithm = ALGO_RSA_DECRYPT;
@@ -936,8 +1042,12 @@ static int sc_hsm_decode_ecdsa_signature(sc_card_t *card,
 		fieldsizebytes = 32;
 	} else if (datalen <= 90) {		// 320 bit curve = 40 * 2 + 10 byte maximum DER signature
 		fieldsizebytes = 40;
-	} else {
+	} else if (datalen <= 106) {		// 384 bit curve = 48 * 2 + 10 byte maximum DER signature
+		fieldsizebytes = 48;
+	} else if (datalen <= 138) {		// 512 bit curve = 64 * 2 + 10 byte maximum DER signature
 		fieldsizebytes = 64;
+	} else {
+		fieldsizebytes = 66;
 	}
 
 	sc_log(card->ctx,
@@ -990,8 +1100,7 @@ static int sc_hsm_compute_signature(sc_card_t *card,
 {
 	int r;
 	sc_apdu_t apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 rbuf[514];
 	sc_hsm_private_data_t *priv;
 
 	if (card == NULL || data == NULL || out == NULL) {
@@ -1003,19 +1112,13 @@ static int sc_hsm_compute_signature(sc_card_t *card,
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OBJECT_NOT_FOUND);
 	}
 
-	// check if datalen exceeds the buffer size
-	if (datalen > SC_MAX_APDU_BUFFER_SIZE) {
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
-	}
-
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0x68, priv->env->key_ref[0], priv->algorithm);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_EXT, 0x68, priv->env->key_ref[0], priv->algorithm);
 	apdu.cla = 0x80;
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof(rbuf);
-	apdu.le = 256;
+	apdu.le = 512;
 
-	memcpy(sbuf, data, datalen);
-	apdu.data = sbuf;
+	apdu.data = data;
 	apdu.lc = datalen;
 	apdu.datalen = datalen;
 	r = sc_transmit_apdu(card, &apdu);
@@ -1045,7 +1148,7 @@ static int sc_hsm_decipher(sc_card_t *card, const u8 * crgram, size_t crgram_len
 	int r;
 	size_t len;
 	sc_apdu_t apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 rbuf[514];
 	sc_hsm_private_data_t *priv;
 
 	if (card == NULL || crgram == NULL || out == NULL) {
@@ -1054,11 +1157,11 @@ static int sc_hsm_decipher(sc_card_t *card, const u8 * crgram, size_t crgram_len
 	LOG_FUNC_CALLED(card->ctx);
 	priv = (sc_hsm_private_data_t *) card->drv_data;
 
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0x62, priv->env->key_ref[0], priv->algorithm);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_EXT, 0x62, priv->env->key_ref[0], priv->algorithm);
 	apdu.cla = 0x80;
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof(rbuf);
-	apdu.le = 256;
+	apdu.le = 512;
 
 	apdu.data = (u8 *)crgram;
 	apdu.lc = crgram_len;
@@ -1130,7 +1233,7 @@ static int sc_hsm_initialize(sc_card_t *card, sc_cardctl_sc_hsm_init_param_t *pa
 	int r;
 	size_t tilen;
 	sc_apdu_t apdu;
-	u8 ibuff[50], *p;
+	u8 ibuff[50+0xFF], *p;
 
 	LOG_FUNC_CALLED(card->ctx);
 
@@ -1140,8 +1243,11 @@ static int sc_hsm_initialize(sc_card_t *card, sc_cardctl_sc_hsm_init_param_t *pa
 	memcpy(p, params->options, 2);
 	p += 2;
 
+	if (params->user_pin_len > 0xFF) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	*p++ = 0x81;	// User PIN
-	*p++ = params->user_pin_len;
+	*p++ = (u8) params->user_pin_len;
 	memcpy(p, params->user_pin, params->user_pin_len);
 	p += params->user_pin_len;
 
@@ -1250,7 +1356,7 @@ static int sc_hsm_wrap_key(sc_card_t *card, sc_cardctl_sc_hsm_wrapped_key_t *par
 {
 	sc_context_t *ctx = card->ctx;
 	sc_apdu_t apdu;
-	u8 data[MAX_EXT_APDU_LENGTH];
+	u8 data[1500];
 	int r;
 
 	LOG_FUNC_CALLED(card->ctx);
@@ -1421,7 +1527,7 @@ static int sc_hsm_init_pin(sc_card_t *card, sc_cardctl_pkcs11_init_pin_t *params
 
 static int sc_hsm_generate_keypair(sc_card_t *card, sc_cardctl_sc_hsm_keygen_info_t *keyinfo)
 {
-	u8 rbuf[1024];
+	u8 rbuf[1200];
 	int r;
 	sc_apdu_t apdu;
 
@@ -1484,12 +1590,12 @@ static int sc_hsm_card_ctl(sc_card_t *card, unsigned long cmd, void *ptr)
 
 static int sc_hsm_init(struct sc_card *card)
 {
-#ifdef _WIN32
+#if defined(ENABLE_OPENPACE) && defined(_WIN32)
 	char expanded_val[PATH_MAX];
 	size_t expanded_len = PATH_MAX;
 #endif
 	int flags,ext_flags;
-	sc_file_t *file;
+	sc_file_t *file = NULL;
 	sc_path_t path;
 	sc_hsm_private_data_t *priv = card->drv_data;
 
@@ -1502,11 +1608,13 @@ static int sc_hsm_init(struct sc_card *card)
 		card->drv_data = priv;
 	}
 
-	flags = SC_ALGORITHM_RSA_RAW|SC_ALGORITHM_ONBOARD_KEY_GEN;
+	flags = SC_ALGORITHM_RSA_RAW|SC_ALGORITHM_RSA_PAD_PSS|SC_ALGORITHM_ONBOARD_KEY_GEN;
 
 	_sc_card_add_rsa_alg(card, 1024, flags, 0);
 	_sc_card_add_rsa_alg(card, 1536, flags, 0);
 	_sc_card_add_rsa_alg(card, 2048, flags, 0);
+	_sc_card_add_rsa_alg(card, 3072, flags, 0);
+	_sc_card_add_rsa_alg(card, 4096, flags, 0);
 
 	flags = SC_ALGORITHM_ECDSA_RAW|
 		SC_ALGORITHM_ECDH_CDH_RAW|
@@ -1525,19 +1633,25 @@ static int sc_hsm_init(struct sc_card *card)
 	_sc_card_add_ec_alg(card, 224, flags, ext_flags, NULL);
 	_sc_card_add_ec_alg(card, 256, flags, ext_flags, NULL);
 	_sc_card_add_ec_alg(card, 320, flags, ext_flags, NULL);
+	_sc_card_add_ec_alg(card, 384, flags, ext_flags, NULL);
+	_sc_card_add_ec_alg(card, 512, flags, ext_flags, NULL);
+	_sc_card_add_ec_alg(card, 521, flags, ext_flags, NULL);
 
 	card->caps |= SC_CARD_CAP_RNG|SC_CARD_CAP_APDU_EXT|SC_CARD_CAP_ISO7816_PIN_INFO;
 
 	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
 	if (sc_hsm_select_file_ex(card, &path, 0, &file) == SC_SUCCESS
-			&& file->prop_attr && file->prop_attr_len >= 5) {
+			&& file && file->prop_attr && file->prop_attr_len >= 2) {
 		static char card_name[SC_MAX_APDU_BUFFER_SIZE];
-		u8 type = file->prop_attr[2];
-		u8 major = file->prop_attr[3];
-		u8 minor = file->prop_attr[4];
+		u8 type = 0xFF;
+		u8 major = file->prop_attr[file->prop_attr_len - 2];
+		u8 minor = file->prop_attr[file->prop_attr_len - 1];
 		char p00[] = "SmartCard-HSM Applet for JCOP";
 		char p01[] = "SmartCard-HSM Demo Applet for JCOP";
 		char *p = "SmartCard-HSM";
+		if (file->prop_attr_len >= 3) {
+			type = file->prop_attr[file->prop_attr_len - 3];
+		}
 		switch (type) {
 			case 0x00:
 				p = p00;
@@ -1554,11 +1668,12 @@ static int sc_hsm_init(struct sc_card *card)
 		if (file->prop_attr[1] & 0x04) {
 			card->caps |= SC_CARD_CAP_SESSION_PIN;
 		}
-		sc_file_free(file);
 	}
+	sc_file_free(file);
 
 	card->max_send_size = 1431;		// 1439 buffer size - 8 byte TLV because of odd ins in UPDATE BINARY
-	if (card->type == SC_CARD_TYPE_SC_HSM_SOC) {
+	if (card->type == SC_CARD_TYPE_SC_HSM_SOC
+			|| card->type == SC_CARD_TYPE_SC_HSM_GOID) {
 		card->max_recv_size = 0x0630;	// SoC Proxy forces this limit
 	} else {
 		card->max_recv_size = 0;		// Card supports sending with extended length APDU and without limit
@@ -1586,7 +1701,9 @@ static int sc_hsm_init(struct sc_card *card)
 static int sc_hsm_finish(sc_card_t * card)
 {
 	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+#ifdef ENABLE_SM
 	sc_sm_stop(card);
+#endif
 	if (priv->serialno) {
 		free(priv->serialno);
 	}
@@ -1596,10 +1713,6 @@ static int sc_hsm_finish(sc_card_t * card)
 	free(priv->EF_C_DevAut);
 	free(priv);
 
-#ifdef ENABLE_OPENPACE
-	EAC_cleanup();
-#endif
-
 	return SC_SUCCESS;
 }
 
@@ -1615,6 +1728,7 @@ static struct sc_card_driver * sc_get_driver(void)
 	sc_hsm_ops                   = *iso_drv->ops;
 	sc_hsm_ops.match_card        = sc_hsm_match_card;
 	sc_hsm_ops.select_file       = sc_hsm_select_file;
+	sc_hsm_ops.get_challenge     = sc_hsm_get_challenge;
 	sc_hsm_ops.read_binary       = sc_hsm_read_binary;
 	sc_hsm_ops.update_binary     = sc_hsm_update_binary;
 	sc_hsm_ops.list_files        = sc_hsm_list_files;
