@@ -39,9 +39,31 @@
 #include "common/libscdl.h"
 #include "internal.h"
 
+static int ignored_reader(sc_context_t *ctx, sc_reader_t *reader)
+{
+	if (ctx != NULL && reader != NULL && reader->name != NULL) {
+		size_t i;
+		const scconf_list *list;
+
+		for (i = 0; ctx->conf_blocks[i]; i++) {
+			list = scconf_find_list(ctx->conf_blocks[i], "ignored_readers");
+			while (list != NULL) {
+				if (strstr(reader->name, list->data) != NULL) {
+					sc_log(ctx, "Ignoring reader \'%s\' because of \'%s\'\n",
+							reader->name, list->data);
+					return 1;
+				}
+				list = list->next;
+			}
+		}
+	}
+
+	return 0;
+}
+
 int _sc_add_reader(sc_context_t *ctx, sc_reader_t *reader)
 {
-	if (reader == NULL) {
+	if (reader == NULL || ignored_reader(ctx, reader)) {
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 	reader->ctx = ctx;
@@ -96,20 +118,23 @@ static const struct _sc_driver_entry internal_card_drivers[] = {
 #endif
 	{ "rutoken",	(void *(*)(void)) sc_get_rutoken_driver },
 	{ "rutoken_ecp",(void *(*)(void)) sc_get_rtecp_driver },
-	{ "westcos",	(void *(*)(void)) sc_get_westcos_driver },
 	{ "myeid",      (void *(*)(void)) sc_get_myeid_driver },
-	{ "sc-hsm",		(void *(*)(void)) sc_get_sc_hsm_driver },
 #if defined(ENABLE_OPENSSL) && defined(ENABLE_SM)
 	{ "dnie",       (void *(*)(void)) sc_get_dnie_driver },
 #endif
 	{ "masktech",	(void *(*)(void)) sc_get_masktech_driver },
+	{ "atrust-acos",(void *(*)(void)) sc_get_atrust_acos_driver },
+	{ "westcos",	(void *(*)(void)) sc_get_westcos_driver },
 
-/* Here should be placed drivers that need some APDU transactions to
- * recognise its cards. */
+/* Here should be placed drivers that need some APDU transactions in the
+ * driver's `match_card()` function. */
+	/* MUSCLE card applet returns 9000 on whatever AID is selected, see
+	 * https://github.com/JavaCardOS/MuscleCard-Applet/blob/master/musclecard/src/com/musclecard/CardEdge/CardEdge.java#L326
+	 * put the muscle driver first to cope with this bug. */
+	{ "muscle",	(void *(*)(void)) sc_get_muscle_driver },
+	{ "sc-hsm",	(void *(*)(void)) sc_get_sc_hsm_driver },
 	{ "mcrd",	(void *(*)(void)) sc_get_mcrd_driver },
 	{ "setcos",	(void *(*)(void)) sc_get_setcos_driver },
-	{ "muscle",	(void *(*)(void)) sc_get_muscle_driver },
-	{ "atrust-acos",(void *(*)(void)) sc_get_atrust_acos_driver },
 	{ "PIV-II",	(void *(*)(void)) sc_get_piv_driver },
 	{ "cac",	(void *(*)(void)) sc_get_cac_driver },
 	{ "itacns",	(void *(*)(void)) sc_get_itacns_driver },
@@ -136,18 +161,15 @@ static const struct _sc_driver_entry old_card_drivers[] = {
 struct _sc_ctx_options {
 	struct _sc_driver_entry cdrv[SC_MAX_CARD_DRIVERS];
 	int ccount;
-	char *forced_card_driver;
 };
 
 
 int
-sc_ctx_win32_get_config_value(char *name_env, char *name_reg, char *name_key,
-		char *out, size_t *out_len)
+sc_ctx_win32_get_config_value(const char *name_env,
+	   	const char *name_reg, const char *name_key,
+		void *out, size_t *out_len)
 {
 #ifdef _WIN32
-	char temp[PATH_MAX + 1];
-	char *value = NULL;
-	DWORD temp_len = PATH_MAX;
 	long rc;
 	HKEY hKey;
 
@@ -155,9 +177,14 @@ sc_ctx_win32_get_config_value(char *name_env, char *name_reg, char *name_key,
 		return SC_ERROR_INVALID_ARGUMENTS;
 
 	if (name_env)   {
-		value = getenv(name_env);
-		if (value)
-			goto done;
+		char *value = value = getenv(name_env);
+		if (value) {
+			if (strlen(value) < *out_len)
+				return SC_ERROR_NOT_ENOUGH_MEMORY;
+			memcpy(out, value, strlen(value));
+			*out_len = strlen(value);
+			return SC_SUCCESS;
+		}
 	}
 
 	if (!name_reg)
@@ -168,36 +195,26 @@ sc_ctx_win32_get_config_value(char *name_env, char *name_reg, char *name_key,
 
 	rc = RegOpenKeyExA(HKEY_CURRENT_USER, name_key, 0, KEY_QUERY_VALUE, &hKey);
 	if (rc == ERROR_SUCCESS) {
-		temp_len = PATH_MAX;
-		rc = RegQueryValueEx( hKey, name_reg, NULL, NULL, (LPBYTE) temp, &temp_len);
-		if ((rc == ERROR_SUCCESS) && (temp_len < PATH_MAX))
-			value = temp;
+		DWORD len = *out_len;
+		rc = RegQueryValueEx(hKey, name_reg, NULL, NULL, out, &len);
 		RegCloseKey(hKey);
-	}
-
-	if (!value) {
-		rc = RegOpenKeyExA( HKEY_LOCAL_MACHINE, name_key, 0, KEY_QUERY_VALUE, &hKey );
 		if (rc == ERROR_SUCCESS) {
-			temp_len = PATH_MAX;
-			rc = RegQueryValueEx( hKey, name_reg, NULL, NULL, (LPBYTE) temp, &temp_len);
-			if ((rc == ERROR_SUCCESS) && (temp_len < PATH_MAX))
-				value = temp;
-			RegCloseKey(hKey);
+			*out_len = len;
+			return SC_SUCCESS;
 		}
 	}
 
-done:
-	if (value) {
-		if (strlen(value) >= *out_len)
-			return SC_ERROR_BUFFER_TOO_SMALL;
-		strcpy(out, value);
-		*out_len = strlen(out);
-		return SC_SUCCESS;
+	rc = RegOpenKeyExA(HKEY_LOCAL_MACHINE, name_key, 0, KEY_QUERY_VALUE, &hKey);
+	if (rc == ERROR_SUCCESS) {
+		DWORD len = *out_len;
+		rc = RegQueryValueEx(hKey, name_reg, NULL, NULL, out, &len);
+		RegCloseKey(hKey);
+		if (rc == ERROR_SUCCESS) {
+			*out_len = len;
+			return SC_SUCCESS;
+		}
 	}
 
-	memset(out, 0, *out_len);
-	*out_len = 0;
-
 	return SC_ERROR_OBJECT_NOT_FOUND;
 #else
 	return SC_ERROR_NOT_SUPPORTED;
@@ -303,12 +320,10 @@ int sc_ctx_log_to_file(sc_context_t *ctx, const char* filename)
 		ctx->debug_file = NULL;
 	}
 
-	if (ctx->reopen_log_file)   {
-		if (!ctx->debug_filename)   {
-			if (!filename)
-				filename = "stderr";
-			ctx->debug_filename = strdup(filename);
-		}
+	if (!ctx->debug_filename)   {
+		if (!filename)
+			filename = "stderr";
+		ctx->debug_filename = strdup(filename);
 	}
 
 	if (!filename)
@@ -327,25 +342,35 @@ int sc_ctx_log_to_file(sc_context_t *ctx, const char* filename)
 	return SC_SUCCESS;
 }
 
+static void
+set_drivers(struct _sc_ctx_options *opts, const scconf_list *list)
+{
+	const char *s_internal = "internal", *s_old = "old";
+	if (list != NULL)
+		del_drvs(opts);
+	while (list != NULL) {
+		if (strcmp(list->data, s_internal) == 0)
+			add_internal_drvs(opts);
+		else if (strcmp(list->data, s_old) == 0)
+			add_old_drvs(opts);
+		else
+			add_drv(opts, list->data);
+		list = list->next;
+	}
+}
 
 static int
 load_parameters(sc_context_t *ctx, scconf_block *block, struct _sc_ctx_options *opts)
 {
 	int err = 0;
 	const scconf_list *list;
-	const char *val, *s_internal = "internal", *s_old = "old";
+	const char *val;
 	int debug;
 #ifdef _WIN32
 	char expanded_val[PATH_MAX];
 	DWORD expanded_len;
 #endif
 
-#ifdef _WIN32
-	ctx->reopen_log_file = 1;
-#else
-	ctx->reopen_log_file = scconf_get_bool(block, "reopen_debug_file", 0);
-#endif
-
 	debug = scconf_get_int(block, "debug", ctx->debug);
 	if (debug > ctx->debug)
 		ctx->debug = debug;
@@ -364,10 +389,6 @@ load_parameters(sc_context_t *ctx, scconf_block *block, struct _sc_ctx_options *
 		sc_ctx_log_to_file(ctx, NULL);
 	}
 
-	if (scconf_get_bool (block, "paranoid-memory",
-				ctx->flags & SC_CTX_FLAG_PARANOID_MEMORY))
-		ctx->flags |= SC_CTX_FLAG_PARANOID_MEMORY;
-
 	if (scconf_get_bool (block, "disable_popups",
 				ctx->flags & SC_CTX_FLAG_DISABLE_POPUPS))
 		ctx->flags |= SC_CTX_FLAG_DISABLE_POPUPS;
@@ -376,25 +397,8 @@ load_parameters(sc_context_t *ctx, scconf_block *block, struct _sc_ctx_options *
 				ctx->flags & SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER))
 		ctx->flags |= SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER;
 
-	val = scconf_get_str(block, "force_card_driver", NULL);
-	if (val) {
-		if (opts->forced_card_driver)
-			free(opts->forced_card_driver);
-		opts->forced_card_driver = strdup(val);
-	}
-
 	list = scconf_find_list(block, "card_drivers");
-	if (list != NULL)
-		del_drvs(opts);
-	while (list != NULL) {
-		if (strcmp(list->data, s_internal) == 0)
-			add_internal_drvs(opts);
-		else if (strcmp(list->data, s_old) == 0)
-			add_old_drvs(opts);
-		else
-			add_drv(opts, list->data);
-		list = list->next;
-	}
+	set_drivers(opts, list);
 
 	return err;
 }
@@ -665,13 +669,14 @@ static void process_config_file(sc_context_t *ctx, struct _sc_ctx_options *opts)
 
 	memset(ctx->conf_blocks, 0, sizeof(ctx->conf_blocks));
 #ifdef _WIN32
-	temp_len = PATH_MAX;
+	temp_len = PATH_MAX-1;
 	r = sc_ctx_win32_get_config_value("OPENSC_CONF", "ConfigFile", "Software\\OpenSC Project\\OpenSC",
 		temp_path, &temp_len);
 	if (r)   {
 		sc_log(ctx, "process_config_file doesn't find opensc config file. Please set the registry key.");
 		return;
 	}
+	temp_path[temp_len] = '\0';
 	conf_path = temp_path;
 #else
 	conf_path = getenv("OPENSC_CONF");
@@ -785,6 +790,7 @@ int sc_context_create(sc_context_t **ctx_out, const sc_context_param_t *parm)
 	sc_context_t		*ctx;
 	struct _sc_ctx_options	opts;
 	int			r;
+	char			*driver;
 
 	if (ctx_out == NULL || parm == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
@@ -841,20 +847,17 @@ int sc_context_create(sc_context_t **ctx_out, const sc_context_param_t *parm)
 		return r;
 	}
 
+	driver = getenv("OPENSC_DRIVER");
+	if (driver) {
+		scconf_list *list = NULL;
+		scconf_list_add(&list, driver);
+		set_drivers(&opts, list);
+		scconf_list_destroy(list);
+	}
+
 	load_card_drivers(ctx, &opts);
 	load_card_atrs(ctx);
 
-	if (!opts.forced_card_driver) {
-		char *driver = getenv("OPENSC_DRIVER");
-		if(driver) {
-			opts.forced_card_driver = strdup(driver);
-		}
-	}
-	if (opts.forced_card_driver) {
-		if (SC_SUCCESS != sc_set_card_driver(ctx, opts.forced_card_driver))
-			sc_log(ctx, "Warning: Could not load %s.", opts.forced_card_driver);
-		free(opts.forced_card_driver);
-	}
 	del_drvs(&opts);
 	sc_ctx_detect_readers(ctx);
 	*ctx_out = ctx;
