@@ -2,7 +2,7 @@
  * pkcs15-sec.c: PKCS#15 cryptography functions
  *
  * Copyright (C) 2001, 2002  Juha Yrjölä <juha.yrjola@iki.fi>
- * Copyrigth (C) 2007        Nils Larsch <nils@larsch.net>
+ * Copyright (C) 2007        Nils Larsch <nils@larsch.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -49,7 +49,7 @@ static int select_key_file(struct sc_pkcs15_card *p15card,
 	/* TODO: Why file_app may be NULL -- at least 3F00 has to be present?
 	 * Check validity of the following assumption. */
 	/* For pkcs15-emulated cards, the file_app may be NULL,
-	 * in that case we allways assume an absolute path */
+	 * in that case we always assume an absolute path */
 	if (!prkey->path.len && prkey->path.aid.len) {
 		/* Private key is a SDO allocated in application DF */
 		path = prkey->path;
@@ -420,12 +420,16 @@ int sc_pkcs15_compute_signature(struct sc_pkcs15_card *p15card,
 
 	/* add the padding bytes (if necessary) */
 	if (pad_flags != 0) {
-		size_t tmplen = sizeof(buf);
+		if (flags & SC_ALGORITHM_RSA_PAD_PSS) {
+			// TODO PSS padding
+		} else {
+			size_t tmplen = sizeof(buf);
 
-		r = sc_pkcs1_encode(ctx, pad_flags, tmp, inlen, tmp, &tmplen, modlen);
-		SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "Unable to add padding");
+			r = sc_pkcs1_encode(ctx, pad_flags, tmp, inlen, tmp, &tmplen, modlen);
+			SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, r, "Unable to add padding");
 
-		inlen = tmplen;
+			inlen = tmplen;
+		}
 	}
 	else if ( senv.algorithm == SC_ALGORITHM_RSA &&
 			(flags & SC_ALGORITHM_RSA_PADS) == SC_ALGORITHM_RSA_PAD_NONE) {
@@ -452,6 +456,15 @@ int sc_pkcs15_compute_signature(struct sc_pkcs15_card *p15card,
 	r = use_key(p15card, obj, &senv, sc_compute_signature, tmp, inlen,
 			out, outlen);
 	LOG_TEST_RET(ctx, r, "use_key() failed");
+
+	/* Some cards may return RSA signature as integer without leading zero bytes */
+	/* Already know outlen >= modlen and r >= 0 */
+	if (obj->type == SC_PKCS15_TYPE_PRKEY_RSA && (unsigned)r < modlen) {
+		memmove(out + modlen - r, out, r);
+		memset(out, 0, modlen - r);
+		r = modlen;
+	}
+
 	sc_mem_clear(buf, sizeof(buf));
 
 	LOG_FUNC_RETURN(ctx, r);
