@@ -50,13 +50,19 @@ static const struct sc_asn1_entry c_asn1_twlabel[] = {
 static const struct sc_asn1_entry c_asn1_algorithm_info[7] = {
 	{ "reference",		SC_ASN1_INTEGER,	SC_ASN1_TAG_INTEGER,	0, NULL, NULL },
 	{ "algorithmPKCS#11",	SC_ASN1_INTEGER,	SC_ASN1_TAG_INTEGER,	0, NULL, NULL },
-	{ "parameters",		SC_ASN1_NULL,		SC_ASN1_TAG_NULL,	0, NULL, NULL },
+	{ "parameters",		SC_ASN1_CHOICE,		0,			0, NULL, NULL },
 	{ "supportedOperations",SC_ASN1_BIT_FIELD,	SC_ASN1_TAG_BIT_STRING,	0, NULL, NULL },
 	{ "objId",		SC_ASN1_OBJECT,		SC_ASN1_TAG_OBJECT,	SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "algRef",		SC_ASN1_INTEGER,	SC_ASN1_TAG_INTEGER,	SC_ASN1_OPTIONAL, NULL, NULL },
 	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
+static const struct sc_asn1_entry c_asn1_algorithm_info_parameters[3] = {
+	{ "PKCS15RSAParameters",SC_ASN1_NULL,		SC_ASN1_TAG_NULL,	0, NULL, NULL },
+	{ "PKCS15ECParameters",	SC_ASN1_OBJECT,		SC_ASN1_TAG_OBJECT,	0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
+};
+
 /*
  * in src/libopensc/types.h SC_MAX_SUPPORTED_ALGORITHMS  defined as 8
  */
@@ -134,9 +140,11 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	u8 preferred_language[3];
 	size_t lang_length = sizeof(preferred_language);
 	struct sc_asn1_entry asn1_supported_algorithms[SC_MAX_SUPPORTED_ALGORITHMS + 1],
-			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
+			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7],
+			asn1_algo_infos_parameters[SC_MAX_SUPPORTED_ALGORITHMS][3];
 	size_t reference_len = sizeof(ti->supported_algos[0].reference);
 	size_t mechanism_len = sizeof(ti->supported_algos[0].mechanism);
+	size_t parameter_len = sizeof(ti->supported_algos[0].parameters);
 	size_t operations_len = sizeof(ti->supported_algos[0].operations);
 	size_t algo_ref_len = sizeof(ti->supported_algos[0].algo_ref);
 
@@ -152,14 +160,22 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	sc_format_asn1_entry(asn1_twlabel, label, &label_len, 0);
 	sc_copy_asn1_entry(c_asn1_profile_indication, asn1_profile_indication);
 
-	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS; ii++)
+	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS; ii++) {
 		sc_copy_asn1_entry(c_asn1_algorithm_info, asn1_algo_infos[ii]);
+		sc_copy_asn1_entry(c_asn1_algorithm_info_parameters,
+			asn1_algo_infos_parameters[ii]);
+	}
 	sc_copy_asn1_entry(c_asn1_supported_algorithms, asn1_supported_algorithms);
 
 	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS; ii++)   {
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 0, &ti->supported_algos[ii].reference, &reference_len, 0);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 1, &ti->supported_algos[ii].mechanism, &mechanism_len, 0);
-		sc_format_asn1_entry(asn1_algo_infos[ii] + 2, NULL, NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos[ii] + 2,
+			asn1_algo_infos_parameters[ii], NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 0,
+			NULL, NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 1,
+			&ti->supported_algos[ii].parameters, &parameter_len, 0);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 3, &ti->supported_algos[ii].operations, &operations_len, 0);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 4, &ti->supported_algos[ii].algo_id, NULL, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 5, &ti->supported_algos[ii].algo_ref, &algo_ref_len, 0);
@@ -191,7 +207,7 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	r = sc_asn1_decode(ctx, asn1_tokeninfo, buf, blen, NULL, NULL);
 	LOG_TEST_RET(ctx, r, "ASN.1 parsing of EF(TokenInfo) failed");
 
-	if (asn1_toki_attrs[1].flags & SC_ASN1_PRESENT)   {
+	if (asn1_toki_attrs[1].flags & SC_ASN1_PRESENT && serial_len > 0)   {
 		ti->serial_number = malloc(serial_len * 2 + 1);
 		if (ti->serial_number == NULL)
 			return SC_ERROR_OUT_OF_MEMORY;
@@ -270,9 +286,11 @@ sc_pkcs15_encode_tokeninfo(sc_context_t *ctx, sc_pkcs15_tokeninfo_t *ti,
 	struct sc_asn1_entry asn1_toki_attrs[C_ASN1_TOKI_ATTRS_SIZE];
 	struct sc_asn1_entry asn1_tokeninfo[2];
 	struct sc_asn1_entry asn1_supported_algorithms[SC_MAX_SUPPORTED_ALGORITHMS + 1],
-			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
+			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7],
+			asn1_algo_infos_parameters[SC_MAX_SUPPORTED_ALGORITHMS][3];
 	size_t reference_len = sizeof(ti->supported_algos[0].reference);
 	size_t mechanism_len = sizeof(ti->supported_algos[0].mechanism);
+	size_t parameter_len = sizeof(ti->supported_algos[0].parameters);
 	size_t operations_len = sizeof(ti->supported_algos[0].operations);
 	size_t algo_ref_len = sizeof(ti->supported_algos[0].algo_ref);
 	struct sc_asn1_entry asn1_last_update[C_ASN1_LAST_UPDATE_SIZE];
@@ -283,14 +301,22 @@ sc_pkcs15_encode_tokeninfo(sc_context_t *ctx, sc_pkcs15_tokeninfo_t *ti,
 	sc_copy_asn1_entry(c_asn1_last_update, asn1_last_update);
 	sc_copy_asn1_entry(c_asn1_profile_indication, asn1_profile_indication);
 
-	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS && ti->supported_algos[ii].reference; ii++)
+	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS && ti->supported_algos[ii].reference; ii++) {
 		sc_copy_asn1_entry(c_asn1_algorithm_info, asn1_algo_infos[ii]);
+		sc_copy_asn1_entry(c_asn1_algorithm_info_parameters,
+			asn1_algo_infos_parameters[ii]);
+	}
 	sc_copy_asn1_entry(c_asn1_supported_algorithms, asn1_supported_algorithms);
 
 	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS && ti->supported_algos[ii].reference; ii++)   {
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 0, &ti->supported_algos[ii].reference, &reference_len, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 1, &ti->supported_algos[ii].mechanism, &mechanism_len, 1);
-		sc_format_asn1_entry(asn1_algo_infos[ii] + 2, NULL, NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos[ii] + 2,
+			asn1_algo_infos_parameters[ii], NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 0,
+			NULL, NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 1,
+			&ti->supported_algos[ii].parameters, &parameter_len, 0);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 3, &ti->supported_algos[ii].operations, &operations_len, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 4, &ti->supported_algos[ii].algo_id, NULL, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 5, &ti->supported_algos[ii].algo_ref, &algo_ref_len, 1);
@@ -406,63 +432,13 @@ fix_authentic_ddo(struct sc_pkcs15_card *p15card)
 	 * Cleanup this attributes -- default values must be OK.
 	 */
 	if (p15card->card->type == SC_CARD_TYPE_OBERTHUR_AUTHENTIC_3_2)   {
-		if (p15card->file_odf != NULL) {
-			sc_file_free(p15card->file_odf);
-			p15card->file_odf = NULL;
-		}
-		if (p15card->file_tokeninfo != NULL) {
-			sc_file_free(p15card->file_tokeninfo);
-			p15card->file_tokeninfo = NULL;
-		}
+		sc_file_free(p15card->file_odf);
+		p15card->file_odf = NULL;
+		sc_file_free(p15card->file_tokeninfo);
+		p15card->file_tokeninfo = NULL;
 	}
 }
 
-
-static void
-fix_starcos_pkcs15_card(struct sc_pkcs15_card *p15card)
-{
-	struct sc_context *ctx = p15card->card->ctx;
-
-	/* set special flags based on card meta data */
-	if (strcmp(p15card->card->driver->short_name,"cardos") == 0
-			&& p15card->tokeninfo && p15card->tokeninfo->label) {
-
-		/* D-Trust cards (D-TRUST, D-SIGN) */
-		if (strstr(p15card->tokeninfo->label,"D-TRUST") != NULL
-				|| strstr(p15card->tokeninfo->label,"D-SIGN") != NULL) {
-
-			/* D-TRUST Card 2.0 2cc (standard cards, which always add
-			 * SHA1 prefix itself */
-			if (strstr(p15card->tokeninfo->label, "2cc") != NULL) {
-				p15card->card->caps |= SC_CARD_CAP_ONLY_RAW_HASH_STRIPPED;
-				sc_log(ctx, "D-TRUST 2cc card detected, only SHA1 works with this card");
-				/* XXX: add detection when other hash than SHA1 is used with
-				 *      such a card, as this produces invalid signatures.
-				 */
-			}
-
-			/* D-SIGN multicard 2.0 2ca (cards working with all types of hashes
-			 * and no addition of prefix) */
-			else if (strstr(p15card->tokeninfo->label, "2ca") != NULL) {
-				p15card->card->caps |= SC_CARD_CAP_ONLY_RAW_HASH;
-				sc_log(ctx, "D-TRUST 2ca card detected");
-			}
-
-			/* D-TRUST card 2.4 2ce (cards working with all types of hashes
-			 * and no addition of prefix) */
-			else if (strstr(p15card->tokeninfo->label, "2ce") != NULL) {
-				p15card->card->caps |= SC_CARD_CAP_ONLY_RAW_HASH;
-				sc_log(ctx, "D-TRUST 2ce card detected");
-			}
-
-			/* XXX: probably there are more D-Trust card in the wild,
-			 *      which also need these flags to produce valid signatures
-			 */
-		}
-	}
-}
-
-
 static int
 parse_ddo(struct sc_pkcs15_card *p15card, const u8 * buf, size_t buflen)
 {
@@ -518,18 +494,12 @@ parse_ddo(struct sc_pkcs15_card *p15card, const u8 * buf, size_t buflen)
 	fix_authentic_ddo(p15card);
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 mem_err:
-	if (p15card->file_odf != NULL) {
-		sc_file_free(p15card->file_odf);
-		p15card->file_odf = NULL;
-	}
-	if (p15card->file_tokeninfo != NULL) {
-		sc_file_free(p15card->file_tokeninfo);
-		p15card->file_tokeninfo = NULL;
-	}
-	if (p15card->file_unusedspace != NULL) {
-		sc_file_free(p15card->file_unusedspace);
-		p15card->file_unusedspace = NULL;
-	}
+	sc_file_free(p15card->file_odf);
+	p15card->file_odf = NULL;
+	sc_file_free(p15card->file_tokeninfo);
+	p15card->file_tokeninfo = NULL;
+	sc_file_free(p15card->file_unusedspace);
+	p15card->file_unusedspace = NULL;
 	LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 }
 
@@ -722,64 +692,72 @@ sc_pkcs15_card_new(void)
 		return NULL;
 	}
 
-	sc_init_oid(&p15card->tokeninfo->profile_indication.oid);
-
 	p15card->magic = SC_PKCS15_CARD_MAGIC;
 	return p15card;
 }
 
 
-void
-sc_pkcs15_free_tokeninfo(struct sc_pkcs15_card *p15card)
+struct sc_pkcs15_tokeninfo *
+sc_pkcs15_tokeninfo_new(void)
 {
-	if (!p15card || !p15card->tokeninfo)
+	struct sc_pkcs15_tokeninfo *tokeninfo;
+
+	tokeninfo = calloc(1, sizeof(struct sc_pkcs15_tokeninfo));
+	if (tokeninfo == NULL) {
+		return NULL;
+	}
+
+	sc_init_oid(&tokeninfo->profile_indication.oid);
+
+	return tokeninfo;
+}
+
+
+void
+sc_pkcs15_free_tokeninfo(struct sc_pkcs15_tokeninfo *tokeninfo)
+{
+	if (!tokeninfo)
 		return;
 
-	if (p15card->tokeninfo->label != NULL)
-		free(p15card->tokeninfo->label);
-	if (p15card->tokeninfo->serial_number != NULL)
-		free(p15card->tokeninfo->serial_number);
-	if (p15card->tokeninfo->manufacturer_id != NULL)
-		free(p15card->tokeninfo->manufacturer_id);
-	if (p15card->tokeninfo->last_update.gtime != NULL)
-		free(p15card->tokeninfo->last_update.gtime);
-	if (p15card->tokeninfo->preferred_language != NULL)
-		free(p15card->tokeninfo->preferred_language);
-	if (p15card->tokeninfo->profile_indication.name != NULL)
-		free(p15card->tokeninfo->profile_indication.name);
-	if (p15card->tokeninfo->seInfo != NULL) {
+	if (tokeninfo->label != NULL)
+		free(tokeninfo->label);
+	if (tokeninfo->serial_number != NULL)
+		free(tokeninfo->serial_number);
+	if (tokeninfo->manufacturer_id != NULL)
+		free(tokeninfo->manufacturer_id);
+	if (tokeninfo->last_update.gtime != NULL)
+		free(tokeninfo->last_update.gtime);
+	if (tokeninfo->preferred_language != NULL)
+		free(tokeninfo->preferred_language);
+	if (tokeninfo->profile_indication.name != NULL)
+		free(tokeninfo->profile_indication.name);
+	if (tokeninfo->seInfo != NULL) {
 		unsigned i;
-		for (i = 0; i < p15card->tokeninfo->num_seInfo; i++)
-			free(p15card->tokeninfo->seInfo[i]);
-		free(p15card->tokeninfo->seInfo);
+		for (i = 0; i < tokeninfo->num_seInfo; i++)
+			free(tokeninfo->seInfo[i]);
+		free(tokeninfo->seInfo);
 	}
-	free(p15card->tokeninfo);
-
-	p15card->tokeninfo = NULL;
+	free(tokeninfo);
 }
 
 
 void
 sc_pkcs15_free_app(struct sc_pkcs15_card *p15card)
 {
-	if (!p15card || !p15card->app)
-		return;
-
-	if (p15card->app->label)
+	if (p15card && p15card->app) {
 		free(p15card->app->label);
-	if (p15card->app->ddo.value)
 		free(p15card->app->ddo.value);
-	free(p15card->app);
-	p15card->app = NULL;
+		free(p15card->app);
+		p15card->app = NULL;
+	}
 }
 
 
 void
 sc_pkcs15_card_free(struct sc_pkcs15_card *p15card)
 {
-	if (p15card == NULL)
+	if (p15card == NULL || p15card->magic != SC_PKCS15_CARD_MAGIC)
 		return;
-	assert(p15card->magic == SC_PKCS15_CARD_MAGIC);
 
 	if (p15card->ops.clear)
 		p15card->ops.clear(p15card);
@@ -794,17 +772,13 @@ sc_pkcs15_card_free(struct sc_pkcs15_card *p15card)
 	sc_pkcs15_free_unusedspace(p15card);
 	p15card->unusedspace_read = 0;
 
-	if (p15card->file_app != NULL)
-		sc_file_free(p15card->file_app);
-	if (p15card->file_tokeninfo != NULL)
-		sc_file_free(p15card->file_tokeninfo);
-	if (p15card->file_odf != NULL)
-		sc_file_free(p15card->file_odf);
-	if (p15card->file_unusedspace != NULL)
-		sc_file_free(p15card->file_unusedspace);
+	sc_file_free(p15card->file_app);
+	sc_file_free(p15card->file_tokeninfo);
+	sc_file_free(p15card->file_odf);
+	sc_file_free(p15card->file_unusedspace);
 
 	p15card->magic = 0;
-	sc_pkcs15_free_tokeninfo(p15card);
+	sc_pkcs15_free_tokeninfo(p15card->tokeninfo);
 	sc_pkcs15_free_app(p15card);
 	free(p15card);
 }
@@ -827,22 +801,14 @@ sc_pkcs15_card_clear(struct sc_pkcs15_card *p15card)
 	sc_pkcs15_remove_dfs(p15card);
 
 	p15card->df_list = NULL;
-	if (p15card->file_app != NULL) {
-		sc_file_free(p15card->file_app);
-		p15card->file_app = NULL;
-	}
-	if (p15card->file_tokeninfo != NULL) {
-		sc_file_free(p15card->file_tokeninfo);
-		p15card->file_tokeninfo = NULL;
-	}
-	if (p15card->file_odf != NULL) {
-		sc_file_free(p15card->file_odf);
-		p15card->file_odf = NULL;
-	}
-	if (p15card->file_unusedspace != NULL) {
-		sc_file_free(p15card->file_unusedspace);
-		p15card->file_unusedspace = NULL;
-	}
+	sc_file_free(p15card->file_app);
+	p15card->file_app = NULL;
+	sc_file_free(p15card->file_tokeninfo);
+	p15card->file_tokeninfo = NULL;
+	sc_file_free(p15card->file_odf);
+	p15card->file_odf = NULL;
+	sc_file_free(p15card->file_unusedspace);
+	p15card->file_unusedspace = NULL;
 	if (p15card->tokeninfo->label != NULL) {
 		free(p15card->tokeninfo->label);
 		p15card->tokeninfo->label = NULL;
@@ -1074,17 +1040,30 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	if(buf == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 
-	err = sc_read_binary(card, 0, buf, len, 0);
-	if (err < 0)   {
-		sc_log(ctx, "read EF(ODF) file error: %s", sc_strerror(err));
-		goto end;
+	err = -1; /* file state: not in cache */
+	if (p15card->opts.use_file_cache) {
+		err = sc_pkcs15_read_cached_file(p15card, &tmppath, &buf, &len);
+		if (err == SC_SUCCESS)
+			err = len;
 	}
-	else if (err < 2) {
-		err = SC_ERROR_PKCS15_APP_NOT_FOUND;
-		sc_log(ctx, "Invalid content of EF(ODF): %s", sc_strerror(err));
-		goto end;
+	if (err < 0) {
+		err = sc_read_binary(card, 0, buf, len, 0);
+		if (err < 2) {
+			if (err < 0) {
+				sc_log(ctx, "read EF(ODF) file error: %s", sc_strerror(err));
+			} else {
+				err = SC_ERROR_PKCS15_APP_NOT_FOUND;
+				sc_log(ctx, "Invalid content of EF(ODF): %s", sc_strerror(err));
+			}
+			goto end;
+		}
+		/* sc_read_binary may return less than requested */
+		len = err;
+
+		if (p15card->opts.use_file_cache) {
+			sc_pkcs15_cache_file(p15card, &tmppath, buf, len);
+		}
 	}
-	len = err;
 
 	if (parse_odf(buf, len, p15card)) {
 		err = SC_ERROR_PKCS15_APP_NOT_FOUND;
@@ -1129,15 +1108,29 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	if(buf == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 
-	err = sc_read_binary(card, 0, buf, len, 0);
-	if (err < 0)   {
-		sc_log(ctx, "read EF(TokenInfo) file error: %s", sc_strerror(err));
-		goto end;
+	err = -1; /* file state: not in cache */
+	if (p15card->opts.use_file_cache) {
+		err = sc_pkcs15_read_cached_file(p15card, &tmppath, &buf, &len);
+		if (err == SC_SUCCESS)
+			err = len;
 	}
-	if (err <= 2) {
-		err = SC_ERROR_PKCS15_APP_NOT_FOUND;
-		sc_log(ctx, "Invalid content of EF(TokenInfo): %s", sc_strerror(err));
-		goto end;
+	if (err < 0) {
+		err = sc_read_binary(card, 0, buf, len, 0);
+		if (err <= 2) {
+			if (err < 0)   {
+				sc_log(ctx, "read EF(TokenInfo) file error: %s", sc_strerror(err));
+			} else {
+				err = SC_ERROR_PKCS15_APP_NOT_FOUND;
+				sc_log(ctx, "Invalid content of EF(TokenInfo): %s", sc_strerror(err));
+			}
+			goto end;
+		}
+		/* sc_read_binary may return less than requested */
+		len = err;
+
+		if (p15card->opts.use_file_cache) {
+			sc_pkcs15_cache_file(p15card, &tmppath, buf, len);
+		}
 	}
 
 	memset(&tokeninfo, 0, sizeof(tokeninfo));
@@ -1149,9 +1142,17 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 
 	*(p15card->tokeninfo) = tokeninfo;
 
+	if (!p15card->tokeninfo->serial_number && 0 == card->serialnr.len) {
+		sc_card_ctl(p15card->card, SC_CARDCTL_GET_SERIALNR, &card->serialnr);
+	}
+
 	if (!p15card->tokeninfo->serial_number && card->serialnr.len)   {
 		char *serial = calloc(1, card->serialnr.len*2 + 1);
 		size_t ii;
+		if (!serial) {
+			err = SC_ERROR_OUT_OF_MEMORY;
+			goto end;
+		}
 
 		for(ii=0;ii<card->serialnr.len;ii++)
 			sprintf(serial + ii*2, "%02X", *(card->serialnr.value + ii));
@@ -1187,7 +1188,9 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "application(aid:'%s')", aid ? sc_dump_hex(aid->value, aid->len) : "empty");
 
-	assert(p15card_out != NULL);
+	if (p15card_out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	p15card = sc_pkcs15_card_new();
 	if (p15card == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
@@ -1244,8 +1247,6 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 			goto error;
 	}
 done:
-	fix_starcos_pkcs15_card(p15card);
-
 	*p15card_out = p15card;
 	sc_unlock(card);
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
@@ -1259,7 +1260,9 @@ error:
 int
 sc_pkcs15_unbind(struct sc_pkcs15_card *p15card)
 {
-	assert(p15card != NULL && p15card->magic == SC_PKCS15_CARD_MAGIC);
+	if (p15card == NULL || p15card->magic != SC_PKCS15_CARD_MAGIC) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	LOG_FUNC_CALLED(p15card->card->ctx);
 	if (p15card->dll_handle)
@@ -1364,26 +1367,16 @@ compare_obj_id(struct sc_pkcs15_object *obj, const struct sc_pkcs15_id *id)
 {
 	void *data = obj->data;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_CERT_X509:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_CERT:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_cert_info *) data)->id, id);
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	case SC_PKCS15_TYPE_PRKEY:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_prkey_info *) data)->id, id);
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_pubkey_info *) data)->id, id);
-	case SC_PKCS15_TYPE_SKEY_DES:
-	case SC_PKCS15_TYPE_SKEY_2DES:
-	case SC_PKCS15_TYPE_SKEY_3DES:
+	case SC_PKCS15_TYPE_SKEY:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_skey_info *) data)->id, id);
-	case SC_PKCS15_TYPE_AUTH_PIN:
-	case SC_PKCS15_TYPE_AUTH_BIO:
-	case SC_PKCS15_TYPE_AUTH_AUTHKEY:
+	case SC_PKCS15_TYPE_AUTH:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_auth_info *) data)->auth_id, id);
 	case SC_PKCS15_TYPE_DATA_OBJECT:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_data_info *) data)->id, id);
@@ -1395,7 +1388,7 @@ compare_obj_id(struct sc_pkcs15_object *obj, const struct sc_pkcs15_id *id)
 static int
 sc_obj_app_oid(struct sc_pkcs15_object *obj, const struct sc_object_id *app_oid)
 {
-	if (obj->type & SC_PKCS15_TYPE_DATA_OBJECT)
+	if ((obj->type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_DATA_OBJECT)
 		return sc_compare_oid(&((struct sc_pkcs15_data_info *) obj->data)->app_oid, app_oid);
 	return 0;
 }
@@ -1407,17 +1400,11 @@ compare_obj_usage(struct sc_pkcs15_object *obj, unsigned int mask, unsigned int
 	void		*data = obj->data;
 	unsigned int	usage;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_PRKEY:
 		usage = ((struct sc_pkcs15_prkey_info *) data)->usage;
 		break;
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		usage = ((struct sc_pkcs15_pubkey_info *) data)->usage;
 		break;
 	default:
@@ -1433,8 +1420,8 @@ compare_obj_flags(struct sc_pkcs15_object *obj, unsigned int mask, unsigned int
 	struct sc_pkcs15_auth_info *auth_info;
 	unsigned int	flags;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_AUTH:
 		auth_info = (struct sc_pkcs15_auth_info *) obj->data;
 		if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
 			return 0;
@@ -1454,17 +1441,14 @@ compare_obj_reference(struct sc_pkcs15_object *obj, int value)
 	void		*data = obj->data;
 	int		reference;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_AUTH:
 		auth_info = (struct sc_pkcs15_auth_info *) obj->data;
 		if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
 			return 0;
 		reference = auth_info->attrs.pin.reference;
 		break;
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	case SC_PKCS15_TYPE_PRKEY:
 		reference = ((struct sc_pkcs15_prkey_info *) data)->key_reference;
 		break;
 	default:
@@ -1479,20 +1463,16 @@ compare_obj_path(struct sc_pkcs15_object *obj, const struct sc_path *path)
 {
 	void *data = obj->data;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_CERT_X509:
-		return sc_compare_path(&((struct sc_pkcs15_cert_info *) data)->path, path);
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_PRKEY:
 		return sc_compare_path(&((struct sc_pkcs15_prkey_info *) data)->path, path);
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		return sc_compare_path(&((struct sc_pkcs15_pubkey_info *) data)->path, path);
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	case SC_PKCS15_TYPE_SKEY:
+		return sc_compare_path(&((struct sc_pkcs15_skey_info *) data)->path, path);
+	case SC_PKCS15_TYPE_CERT:
+		return sc_compare_path(&((struct sc_pkcs15_cert_info *) data)->path, path);
+	case SC_PKCS15_TYPE_AUTH:
 		return sc_compare_path(&((struct sc_pkcs15_auth_info *) data)->path, path);
 	case SC_PKCS15_TYPE_DATA_OBJECT:
 		return sc_compare_path(&((struct sc_pkcs15_data_info *) data)->path, path);
@@ -1958,7 +1938,9 @@ sc_pkcs15_encode_df(struct sc_context *ctx, struct sc_pkcs15_card *p15card, stru
 		     unsigned char **nbuf, size_t *nbufsize) = NULL;
 	int r;
 
-	assert(p15card != NULL && p15card->magic == SC_PKCS15_CARD_MAGIC);
+	if (p15card == NULL || p15card->magic != SC_PKCS15_CARD_MAGIC) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	switch (df->type) {
 	case SC_PKCS15_PRKDF:
 		func = sc_pkcs15_encode_prkdf_entry;
@@ -1967,6 +1949,9 @@ sc_pkcs15_encode_df(struct sc_context *ctx, struct sc_pkcs15_card *p15card, stru
 	case SC_PKCS15_PUKDF_TRUSTED:
 		func = sc_pkcs15_encode_pukdf_entry;
 		break;
+	case SC_PKCS15_SKDF:
+		func = sc_pkcs15_encode_skdf_entry;
+		break;
 	case SC_PKCS15_CDF:
 	case SC_PKCS15_CDF_TRUSTED:
 	case SC_PKCS15_CDF_USEFUL:
@@ -2221,7 +2206,8 @@ sc_pkcs15_encode_unusedspace(struct sc_context *ctx, struct sc_pkcs15_card *p15c
 		sc_format_asn1_entry(asn1_unusedspace + c, asn1_values + 3*c, NULL, 1);
 		sc_copy_asn1_entry(c_asn1_unusedspace_values, asn1_values + 3*c);
 		sc_format_asn1_entry(asn1_values + 3*c, &unusedspace->path, NULL, 1);
-		sc_format_asn1_entry(asn1_values + 3*c+1, &unusedspace->auth_id, NULL, unusedspace->auth_id.len);
+		sc_format_asn1_entry(asn1_values + 3*c+1, &unusedspace->auth_id, NULL,
+			   unusedspace->auth_id.len > 0 ? 1 : 0);
 		c++;
 	}
 	asn1_unusedspace[c].name = NULL;
@@ -2300,13 +2286,16 @@ int
 sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_path,
 		unsigned char **buf, size_t *buflen)
 {
-	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_context *ctx;
 	struct sc_file *file = NULL;
 	unsigned char *data = NULL;
 	size_t	len = 0, offset = 0;
 	int	r;
 
-	assert(p15card != NULL && in_path != NULL && buf != NULL);
+	if (p15card == NULL || p15card->card == NULL || in_path == NULL || buf == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	ctx = p15card->card->ctx;
 
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "path=%s, index=%u, count=%d", sc_print_path(in_path), in_path->index, in_path->count);
@@ -2326,7 +2315,8 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 
 	if (r) {
 		r = sc_lock(p15card->card);
-		LOG_TEST_RET(ctx, r, "sc_lock() failed");
+		if (r)
+			goto fail;
 		r = sc_select_file(p15card->card, in_path, &file);
 		if (r)
 			goto fail_unlock;
@@ -2367,7 +2357,6 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 				if (r == SC_ERROR_RECORD_NOT_FOUND)
 					break;
 				if (r < 0) {
-					free(data);
 					goto fail_unlock;
 				}
 				if (r < 2)
@@ -2389,7 +2378,6 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 		else {
 			r = sc_read_binary(p15card->card, offset, data, len, 0);
 			if (r < 0) {
-				free(data);
 				goto fail_unlock;
 			}
 			/* sc_read_binary may return less than requested */
@@ -2398,15 +2386,20 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 		sc_unlock(p15card->card);
 
 		sc_file_free(file);
+
+		if (len && p15card->opts.use_file_cache) {
+			sc_pkcs15_cache_file(p15card, in_path, data, len);
+		}
 	}
 	*buf = data;
 	*buflen = len;
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 
 fail_unlock:
-	if (file)
-		sc_file_free(file);
 	sc_unlock(p15card->card);
+fail:
+	free(data);
+	sc_file_free(file);
 	LOG_FUNC_RETURN(ctx, r);
 }
 
@@ -2414,7 +2407,8 @@ fail_unlock:
 int
 sc_pkcs15_compare_id(const struct sc_pkcs15_id *id1, const struct sc_pkcs15_id *id2)
 {
-	assert(id1 != NULL && id2 != NULL);
+	if (id1 == NULL || id2 == NULL)
+		return 0;
 	if (id1->len != id2->len)
 		return 0;
 	return memcmp(id1->value, id2->value, id1->len) == 0;
@@ -2527,7 +2521,7 @@ sc_pkcs15_allocate_object_content(struct sc_context *ctx, struct sc_pkcs15_objec
 	/* Need to pass by temporary variable,
 	 * because 'value' and 'content.value' pointers can be the sames.
 	 */
-	tmp_buf = (unsigned char *)sc_mem_alloc_secure(ctx, len);
+	tmp_buf = calloc(sizeof *tmp_buf, len);
 	if (!tmp_buf)
 		return SC_ERROR_OUT_OF_MEMORY;
 
@@ -2608,13 +2602,16 @@ sc_pkcs15_add_supported_algo_ref(struct sc_pkcs15_object *obj, struct sc_support
 	if (!algo)
 		return SC_SUCCESS;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_PRKEY_RSA:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_PRKEY:
 		algo_refs = ((struct sc_pkcs15_prkey_info *)obj->data)->algo_refs;
 		break;
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
+	case SC_PKCS15_TYPE_PUBKEY:
 		algo_refs = ((struct sc_pkcs15_pubkey_info *)obj->data)->algo_refs;
 		break;
+	case SC_PKCS15_TYPE_SKEY:
+		algo_refs = ((struct sc_pkcs15_skey_info *)obj->data)->algo_refs;
+		break;
 	}
 	if (!algo_refs)
 		return SC_ERROR_NOT_SUPPORTED;
@@ -2640,23 +2637,20 @@ sc_pkcs15_get_object_id(const struct sc_pkcs15_object *obj, struct sc_pkcs15_id
 	if (!obj || !out)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_CERT_X509:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_CERT:
 		*out = ((struct sc_pkcs15_cert_info *) obj->data)->id;
 		break;
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	case SC_PKCS15_TYPE_PRKEY:
 		*out = ((struct sc_pkcs15_prkey_info *) obj->data)->id;
 		break;
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		*out = ((struct sc_pkcs15_pubkey_info *) obj->data)->id;
 		break;
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	case SC_PKCS15_TYPE_SKEY:
+		*out = ((struct sc_pkcs15_skey_info *) obj->data)->id;
+		break;
+	case SC_PKCS15_TYPE_AUTH:
 		*out = ((struct sc_pkcs15_auth_info *) obj->data)->auth_id;
 		break;
 	case SC_PKCS15_TYPE_DATA_OBJECT:
