@@ -477,7 +477,7 @@ int main(int argc, char *argv[])
 			file = sc_file_new();
 			if(file == NULL)
 			{
-				printf("Not enougth memory.\n");
+				printf("Not enough memory.\n");
 				goto out;
 			}
 
@@ -596,7 +596,7 @@ int main(int argc, char *argv[])
 
 		if(rsa == NULL || bn == NULL || mem == NULL)
 		{
-			printf("Not enougth memory.\n");
+			printf("Not enough memory.\n");
 			goto out;
 		}
 
@@ -608,7 +608,7 @@ int main(int argc, char *argv[])
 
 		if(mem == NULL)
 		{
-			printf("Not enougth memory.\n");
+			printf("Not enough memory.\n");
 			goto out;
 		}
 
@@ -638,7 +638,7 @@ int main(int argc, char *argv[])
 			file = sc_file_new();
 			if(file == NULL)
 			{
-				printf("Not enougth memory.\n");
+				printf("Not enough memory.\n");
 				goto out;
 			}
 
@@ -814,7 +814,7 @@ int main(int argc, char *argv[])
 		b = malloc(file->size);
 		if(b == NULL)
 		{
-				printf("Not enougth memory.\n");
+				printf("Not enough memory.\n");
 				goto out;
 		}
 
@@ -864,14 +864,18 @@ int main(int argc, char *argv[])
 		b = malloc(file->size);
 		if(b == NULL)
 		{
-				printf("Not enougth memory.\n");
-				goto out;
+			printf("Not enough memory.\n");
+			goto out;
 		}
 
 		memset(b, 0, file->size);
 
 		fp = fopen(put_filename, "rb");
-		fread(b, 1, file->size, fp);
+		if (fp == NULL || file->size != fread(b, 1, file->size, fp))
+		{
+			printf("could not read %s.\n", put_filename);
+			goto out;
+		}
 		fclose(fp);
 
 		r = sc_update_binary(card, 0, b, file->size, 0);
